Require Import base.
Require Import ut_term.
Require Import ut_red.
Require Import ut_env.
Require Import ut_typ.
Require Import List.

Open Scope UT_scope.

(* How to deal with Beta reduction in the context *)
Inductive Beta_env : Env -> Env -> Prop :=
 | Beta_env_hd : forall A B e, A → B -> Beta_env (A::e) (B::e)
 | Beta_env_cons : forall A e f, Beta_env e f -> Beta_env (A::e) (A::f)
.

Inductive Betas_env : Env -> Env -> Prop :=
 | Betas_env_refl : forall e, Betas_env e e
 | Betas_env_Beta : forall e f, Beta_env e f -> Betas_env e f
 | Betas_env_trans : forall e f g, Betas_env e f -> Betas_env f g -> Betas_env e g.

Inductive Betac_env : Env -> Env -> Prop :=
 | Betac_env_Betas : forall e f, Betas_env e f -> Betac_env e f
 | Betac_env_sym :forall e f, Betac_env e f -> Betac_env f e
 | Betac_env_trans : forall e f g, Betac_env e f -> Betac_env f g -> Betac_env e g.


Hint Constructors Beta_env Betas_env Betac_env.

Notation " a →e b " := (Beta_env a b) (at level 20) : UT_scope.
Notation " a →→e b " := (Betas_env a b) (at level 20) : UT_scope.
Notation " a ≡e b " := (Betac_env a b) (at level 20) : UT_scope.

(* simples facts on env / beta *)
Lemma Betas_env_nil_ : forall e g,e →→e g -> (e = nil -> g = nil).
intros e g H.
induction H; intros.
trivial.
subst.
inversion H.
intuition.
Qed.

Lemma Betas_env_nil : forall g, nil →→e g -> g = nil.
intros.
apply (Betas_env_nil_ nil g H); intuition.
Qed.

Lemma Betas_env_nil2_ : forall e g , e →→e g -> (g = nil -> e = nil).
intros e g H; induction H; intros.
trivial.
subst; inversion H.
intuition.
Qed.

Lemma Betas_env_nil2 : forall e, e →→e nil -> e = nil.
intros ; apply (Betas_env_nil2_ e nil) ; trivial.
Qed.

Lemma Betas_env_inv_ : forall E F, E →→e F -> 
 (forall a b g f,E = a::g -> F = b::f  -> a→→b /\ g→→e f).
intros E F H; induction H; intros.
rewrite H0 in H.
injection H; intros; subst.
split; constructor.
inversion H; subst.
injection H3; injection H4; intros; subst.
rewrite H5; split; constructor; trivial.
injection H3; injection H4; intros; subst.
rewrite H6; split; constructor; trivial.
destruct f.
rewrite (Betas_env_nil g H0) in *; discriminate.
subst.
destruct (IHBetas_env2 t b f f0); intuition; 
destruct (IHBetas_env1 a t g0 f); intuition.
apply Betas_trans with (y:=t); trivial.
apply Betas_env_trans with  (f:=f); trivial.
Qed.

Lemma Betas_env_inv :forall a b g f, (a::g)  →→e  (b::f) -> a →→ b /\ g→→e f.
intros.
apply Betas_env_inv_ with (a::g) (b::f) ; trivial.
Qed.

Lemma Betas_env_hd : forall A C f, A→→C -> (A::f) →→e (C::f).
intros.
induction H.
constructor.
constructor; apply Beta_env_hd; trivial.
apply Betas_env_trans with (f:=y::f); intuition.
Qed.

Lemma Betas_env_cons : forall A g f , g →→e f -> (A::g) →→e (A::f).
intros.
induction H.
constructor.
constructor; apply Beta_env_cons; trivial.
apply Betas_env_trans with (f:=A::f); trivial.
Qed.

Lemma Betas_env_comp : forall A B g f, A→→B -> g→→e f -> (A::g) →→e (B::f).
intros.
apply Betas_env_trans with (f:= A::f).
apply Betas_env_cons; trivial.
apply Betas_env_hd; trivial.
Qed.


(* Important one 
we have to proof subject reduction and context reduction at the same time
*)

Lemma Beta_env_sound :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ' ⊣ -> Γ →e Γ' -> Γ' ⊢ x : A  .
induction 1; intros.
(**)
constructor; trivial.
(**)
generalize dependent A. revert H1 H v. 
induction H2; intros. destruct H2 as (AA&  ? & ?). inversion H3; subst; clear H3.
inversion H1; subst; clear H1. inversion H0; subst; clear H0.
apply Cnv with ( B  ↑ 1 ) s0. eauto.
constructor. eauto.  exists B; intuition.
change !s0 with !s0 ↑ 1. apply thinning with s ;trivial.
constructor. eauto. exists AA; intuition.
destruct H0 as (AA&  ? & ?).  inversion H3; subst; clear H3.
constructor. trivial. exists A; intuition.
change (S(S n)) with (1+S n). change #(S n) with (#n  ↑ 1).
rewrite <- lift_lift.
inversion H1; subst; clear H1.
apply thinning with s; trivial. apply IHBeta_env. eauto. inversion H; subst; clear H. eauto. 
exists AA; intuition.

econstructor; eauto.
econstructor. apply H. eauto. eauto. eauto.
econstructor. eauto. eauto.

eauto.
Qed.



Lemma Beta_env_sound_up :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ' ⊣ -> Γ' →e Γ -> Γ' ⊢ x : A  .
induction 1; intros.
(**)
constructor; trivial.
(**)
generalize dependent A. revert H1 H v. 
induction H2; intros. destruct H2 as (AA&  ? & ?). inversion H3; subst; clear H3.
inversion H1; subst; clear H1. inversion H0; subst; clear H0.
apply Cnv with ( A  ↑ 1 ) s0. eauto.
constructor. eauto.  exists A; intuition.
change !s0 with !s0 ↑ 1. apply thinning with s ;trivial.
constructor. eauto. exists AA; intuition.
destruct H0 as (AA&  ? & ?).  inversion H3; subst; clear H3.
constructor. trivial. exists A; intuition.
change (S(S n)) with (1+S n). change #(S n) with (#n  ↑ 1).
rewrite <- lift_lift.
inversion H1; subst; clear H1.
apply thinning with s; trivial. apply IHBeta_env. eauto. inversion H; subst; clear H. eauto. 
exists AA; intuition.

econstructor; eauto.
econstructor. apply H. eauto. eauto. eauto.
econstructor. eauto. eauto.

eauto.
Qed.

Lemma Beta_sound : forall Γ x A, Γ ⊢ x : A-> forall y, x→ y ->  Γ ⊢y : A.
induction 1; intros.
(*1*)
inversion H1.
(*2*)
inversion H1.
(*3*)
inversion H2; subst; clear H2. econstructor; eauto. econstructor. apply H. intuition.
eapply Beta_env_sound. apply H1. eauto. intuition.
(*4*)
inversion H3; subst; clear H3. eauto.
apply Cnv with ( Π (B0), B) s3. eauto.
econstructor. apply H. eauto.
eapply Beta_env_sound. apply H1. econstructor. eauto. intuition.
eapply Beta_env_sound. apply H2. econstructor. eauto. intuition.
econstructor; eauto.
(*5*)
inversion H1; subst; clear H1.
assert (exists u, Γ ⊢ Π(A),B : !u). apply TypeCorrect in H.
destruct H. destruct H; discriminate. trivial.
destruct H1. apply Geniii in H1 as (s & t & u & h); decompose [and] h; clear h.
apply Geniv in H. destruct H as (u1 & u2 & u3 & D & ? & ? & ? & ? & ?).
apply Betac_Pi_inv in H as (? & ?).
eapply Cnv with (s := t). eapply Betac_subst2. apply Betac_sym. apply H9.
eapply substitution. apply H7. eapply Cnv. apply H. apply H0. apply H6. constructor.
eauto. change !t with (!t  [ ← b] ). eapply substitution. apply H5. apply H0. constructor.
eauto.

econstructor.  eauto. trivial.
assert (exists u, Γ ⊢ Π(A),B : !u). apply TypeCorrect in H.
destruct H. destruct H; discriminate. trivial.
destruct H1. apply Geniii in H1 as (s & t & u & h); decompose [and] h; clear h.
apply Cnv with (B [← y0]) t. apply Betac_subst. auto.
econstructor. apply H. eauto.
change !t with (!t  [ ← b] ). eapply substitution. apply H6. apply H0. constructor.
eauto.
(**)
eauto.
Qed.


Lemma SubjectRed : forall Γ x T, Γ ⊢ x : T-> 
  forall y , x →→ y -> Γ ⊢ y : T.
intros. induction H0.
trivial. eapply Beta_sound.
apply H. trivial. intuition.
Qed.

(* Type reduction *)
Lemma Betas_typ_sound : forall Γ x T, Γ ⊢ x : T -> 
 forall T', T →→ T' -> Γ ⊢ x : T'.
intros.
assert ((exists s, T = !s )\/ exists s,  Γ ⊢ T : !s ) by ( apply TypeCorrect in H; intuition).
destruct H1 as [[s Hs]|[s Hs]].
rewrite Hs in *. apply Betas_S in H0. rewrite H0; trivial.
apply Cnv with T s; intuition.
apply SubjectRed with T; intuition.
Qed.

Lemma Beta_env_sound2 : forall Γ Γ', Γ ⊣ -> Γ →e Γ' -> Γ' ⊣.
intros. induction H0; intuition.
apply wf_inv in H. destruct H as (s & ? ).
econstructor. eapply Beta_sound. apply H. trivial.
inversion H; subst; clear H.
econstructor. eapply Beta_env_sound. apply H2.  apply IHBeta_env.
eauto. trivial.
Qed.

Lemma Betas_env_sound2 : forall Γ Γ', Γ ⊣ -> Γ →→e Γ' -> Γ' ⊣.
intros.
induction H0; intuition.
eapply Beta_env_sound2. apply H. trivial.
Qed.

Lemma Betas_env_sound :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ →→e Γ' -> Γ' ⊢ x : A  .
intros.
induction H0; intuition.
eapply Beta_env_sound. apply H. apply Beta_env_sound2 with e. apply wf_typ in H. trivial.
trivial. trivial.
Qed.

Lemma Betas_env_sound_up :   forall Γ x A, Γ ⊢ x : A-> forall Γ', Γ' ⊣ -> Γ' →→e Γ -> Γ' ⊢ x : A  .
intros. revert x A H H0 .
induction H1; intros.
trivial.
eapply Beta_env_sound_up. apply H0. trivial. trivial.
apply IHBetas_env1. apply IHBetas_env2. trivial. apply Betas_env_sound2 with e; trivial.
trivial.
Qed.




(* compilation of all this *)
Lemma Subject_Reduction :  forall Γ x A, Γ ⊢ x : A ->
  forall Γ' y B, x→→ y -> Γ →→e Γ' -> A→→ B -> Γ' ⊢ y : B.
intros.
apply Betas_typ_sound with A; intuition.
apply Betas_env_sound with Γ; intuition.
apply SubjectRed with x; intuition.
Qed.


Inductive Tv : Term -> Prop :=
 | Tv_intro : forall x, Tv #x
 | Tv_la: forall b A, Tv b -> Tv (λ[A],b)
 | Tv_app :forall b c, Tv b -> Tv (b·c)
.

Inductive Ts : Term -> Prop :=
 | Ts_intro : forall s, Ts !s
 | Ts_pi : forall A B, Ts (Π(A),B)
 | Ts_la : forall b A, Ts b -> Ts (λ[A],b)
 | Ts_app : forall b c, Ts b -> Ts (b·c)
.

Hint Constructors Tv Ts.

Lemma Tx_ok : forall t, Tv t \/ Ts t.
induction t; intros.
left; trivial.
right; trivial.
destruct IHt1.
left; constructor; trivial.  right; constructor; trivial.
right; trivial.
destruct IHt2.
left; constructor; trivial.  right; constructor; trivial.
Qed.

Lemma fun_item_lift : forall A A' v Γ , A ↓ v ⊂ Γ -> A' ↓ v ⊂ Γ -> A = A'.
intros.
destruct H as (x & ?& ?). destruct H0 as (x' & ?& ?).
subst. replace x' with x. trivial. eapply fun_item. apply H1. apply H2.
Qed.

Lemma TU1 : forall Γ M A , Γ ⊢ M : A -> forall A',  Γ ⊢ M : A' -> Tv M ->  A ≡ A'.
induction 1; intros.
inversion H2.
(**)
apply Genii in H1 as (B & ? &?).
replace A with B. eauto. eapply fun_item_lift. apply H3. trivial.
(**)
inversion H3; subst; clear H3.
(**)
inversion H4; subst; clear H4.
apply Geniv in H3 as (t1 & t2 & t3 & D & h); decompose [and] h;clear h.
eapply Betac_trans. eapply Betac_Pi. apply Betac_refl. eapply IHtyp3. apply H7. trivial.
intuition.
(**)
inversion H2; subst; clear H2.
apply Genv in H1 as (C & D & ? & ?& ?).
destruct (Betac_Pi_inv A B D C). apply IHtyp1. trivial. trivial.
eapply Betac_trans. eapply Betac_subst2. apply H6. intuition.
(**)
eauto.
Qed.

Fixpoint Tels (l : Env) (t : Term) {struct l } : Term :=
 match l with
 | nil => t
 | cons A q => Π(A),(Tels q t)
end.


Lemma Betas_to_Pi :forall A B T, Π(A),B  →→ T ->  exists C, exists D,
  T = Π(C),D /\ A  →→ C /\ B  →→ D.
intros. remember (Π(A),B) as P.
generalize dependent A. generalize dependent B.
induction H;intros;  subst; try discriminate.
exists A; exists B; intuition.
inversion H; subst; clear H.
exists A; exists y0; eauto.
exists B0; exists B; eauto.
destruct (IHBetas1  B A ) as (K & L & -> & ? & ?); trivial.
destruct (IHBetas2 L K ) as (K' & L' & -> & ? & ?); trivial.
exists K'; exists L'; eauto.
Qed.


Lemma foo0 : forall l A B  t, B →→ Tels l t -> Π (A), B →→ Tels (A::l) t.
induction l; intros; simpl in *. 
eauto.
eauto.
Qed.

Lemma Beta_Tels_inv : forall l t T, Tels l t → T -> (exists l', T = Tels l' t /\ l →e l') \/ 
  (exists t', T = Tels l t' /\ t → t').
induction l; intros; simpl in *.
right. exists T; intuition.
inversion H; subst; clear H.
destruct (IHl t y H3). destruct H as (l' & -> & ?).
left. exists (a::l'); intuition.
destruct H as (t' & -> & ?).
right. exists t'; intuition.

left. exists (B::l); intuition.
Qed.


Lemma Betas_Tels_inv : forall l t T, Tels l t →→ T -> exists l', exists t', T = Tels l' t' /\ l →→e l' /\ t →→ t'.
intros.
remember (Tels l t) as TT. revert l t HeqTT.
induction H; intros; subst.
exists l; exists t; intuition.
apply Beta_Tels_inv in H.
destruct H. destruct H as (l' & -> & ?).
exists l'; exists t; intuition.
destruct H as (t' & -> & ?).
exists l; exists t'; intuition.
destruct (IHBetas1 l t) as (l' & t' & -> & ?& ?); trivial.
destruct (IHBetas2 l' t') as (l'' & t'' & -> & ?& ?); trivial.
exists l''; exists t''; eauto.
Qed.


Lemma Betas_Tels : forall l l' t t', l →→e l' -> t →→ t' -> Tels l t →→ Tels l' t'.
assert (forall l l', l →→e l' -> forall t ,  Tels l t →→ Tels l' t).
induction 1; intros; eauto.
induction H; simpl; eauto.
assert (forall t t', t →→ t' -> forall l ,  Tels l t →→ Tels l t').
induction 1; intros; eauto.
induction l; simpl; eauto.
intros.
eauto.
Qed.

Lemma Betas_not_Pi_to_Sort : forall A B s,  ~ (Π (A), B →→ !s).
intros; intro.
apply Betas_to_Pi in H as (C & D & ? & ?& ?); discriminate.
Qed.


Fixpoint env_subst E n N {struct E } := match E with
 | nil => nil 
 | A::F => (A[n ← N]) :: (env_subst F (S n) N)
end.

Notation "E  [[ n ← N ]]" := (env_subst E n N) (at level 80).

Lemma Tels_subst : forall l t n N, (Tels l t) [n ← N] = Tels (l [[n ← N]]) (t [(length l + n) ← N]).
induction l; simpl; intros.
trivial.
rewrite IHl.
rewrite plus_n_Sm. trivial.
Qed.

Lemma TU2 : forall Γ M A,  Γ ⊢ M : A -> forall A', Γ ⊢ M : A' -> Ts M -> 
  exists s, exists s', exists l,  A →→ (Tels l !s) /\ A' →→ (Tels l !s').
induction 1; intros.
apply Geni in H1 as (s' & ?& ?).
apply Betac_sym in H1. apply conv_to_sort in H1.
exists t; exists s'; exists nil; simpl; intuition.
(**)
inversion H2.
(**)
clear IHtyp1 IHtyp2.
apply Geniii in H2 as (s1 & s2 & s3 & h); decompose [and] h; clear h.
apply Betac_sym in H2. apply conv_to_sort in H2.
exists u; exists s3; exists nil; simpl; intuition.
(**)
inversion H4; subst; clear H4.
apply Geniv in H3 as (t1 & t2 & t3 & D & h); decompose [and] h; clear h.
destruct (IHtyp3 D H7 H6) as (u & v & l & ?& ?).
apply Betac_confl in H3 as (Z & ? &?).
apply Betas_to_Pi in H11 as (K & L & -> & ? & ?).
destruct (Betas_diamond  D L (Tels l !v)) as ( Z & ?& ?); trivial.
apply Betas_Tels_inv in H14 as (l'' & ? & -> & ? &?).
apply Betas_S in H15; subst.
exists u; exists v; exists (K::l''); split.
apply Betas_trans with (Tels (A::l) !u). simpl. eauto. simpl.
apply Betas_Pi. trivial. apply Betas_Tels; trivial.
simpl. eauto.
(**)
inversion H2; subst; clear H2. clear IHtyp2.
apply Genv in H1 as (C & D & ? & ?&  ?). destruct (IHtyp1 ( Π (D),C) ) as (u & v & l & ? & ?); trivial.
clear IHtyp1. destruct l; simpl in *. apply Betas_not_Pi_to_Sort in H5; elim H5.
apply Betas_to_Pi in H5 as (B0 & A0 & ? & ?& ?). injection H5; intros;subst; clear H5.
apply Betas_to_Pi in H6 as (D0 & C0 & ? & ?& ?). injection H5; intros;subst; clear H5.
apply Betac_confl in H1 as (Z & ?& ?).
destruct (Betas_diamond (C[←b]) Z ((Tels l !v)[←b]) H5) as (ZZ & ?& ?).
apply Betas_subst2; trivial.  rewrite Tels_subst in H11. simpl in H11.
apply Betas_Tels_inv in H11. destruct H11 as (l' & x & -> & ?& ?). apply Betas_S in H12. subst.
exists u ; exists v; exists l'; split. eapply Betas_trans. apply Betas_subst2.  apply H8.
rewrite Tels_subst. eapply Betas_Tels. trivial. simpl; trivial.
eauto.
(**)
destruct (IHtyp1 A' H2 H3) as (u & v & l & ? & ?).
apply Betac_confl in H as (Z & ? & ?). destruct (Betas_diamond A (Tels l !u) Z H4 H) as (ZZ & ?& ?).
apply Betas_Tels_inv in H7 as (l' & ? & -> & ? & ?). apply Betas_S in H9; subst.
exists u; exists v; exists l'; split.
apply Betas_trans with Z; trivial. apply Betas_trans with (Tels l !v); trivial.
apply Betas_Tels; trivial.
Qed.

Lemma Betas_Tels_length : forall l l' s t, Tels l !s →→ Tels l' !t -> length l = length l'.
induction l; destruct l'; simpl in *; intros.
trivial. apply Betas_S in H; discriminate. apply Betas_Pi_inv in H as (C & D & ? & ?& ?); discriminate.
apply Betas_Pi_inv in H as (C & D & ?&  ? & ?). injection H; intros; subst; clear H.
rewrite (IHl l' s t0 H1). trivial.
Qed.


Lemma Tels_inj : forall l l' s t, Tels l !s = Tels l' !t -> l = l'.
induction l; destruct l'; simpl in *; intros.
trivial. discriminate. discriminate. injection H; intros; subst; clear H.
rewrite (IHl l' s t0 H0); trivial.
Qed.



Lemma weak_tsu_env : forall Γ M l s , Γ ⊢ M : Tels l !s -> Ts M -> forall l' t, Γ ⊢ M : Tels l' !t -> l ≡e l'.
intros. destruct (TU2 Γ M (Tels l !s) H (Tels l' !t) H1 H0) as ( u & v &  k & ? & ?).
apply Betas_Tels_inv in H2. destruct H2 as (k' & t' & ? & ?& ?). apply Betas_S in H5; subst.
apply Betas_Tels_inv in H3. destruct H3 as (k'' & s' & ? & ?& ?). apply Betas_S in H6; subst.
apply Tels_inj in H2. apply Tels_inj in H3. subst. rewrite H3 in H4. eauto.
Qed.



